use std::time::Duration;

use color_eyre::eyre::Report;
use oauth2::{
    basic::BasicClient, reqwest::async_http_client, AuthUrl, ClientId, DeviceAuthorizationUrl,
    Scope, StandardDeviceAuthorizationResponse, TokenResponse, TokenUrl,
};
use reqwest::{Response, StatusCode, Url};
use serde::Deserialize;
use serde_json::json;
use thiserror::Error;

const CLIENT_ID: &str = "39bd526b-1897-42c9-b652-8b987550ab89";

const DEVICE_CODE_URL: &str = "https://login.microsoftonline.com/consumers/oauth2/v2.0/devicecode";
const MSA_AUTHORIZE_URL: &str = "https://login.microsoftonline.com/consumers/oauth2/v2.0/authorize";
const MSA_TOKEN_URL: &str = "https://login.microsoftonline.com/common/oauth2/v2.0/token";
const XBOX_USER_AUTH_URL: &str = "https://user.auth.xboxlive.com/user/authenticate";
const XBOX_XSTS_AUTHORIZE_URL: &str = "https://xsts.auth.xboxlive.com/xsts/authorize";
const MINECRAFT_AUTH_URL: &str = "https://api.minecraftservices.com/authentication/login_with_xbox";
const MINECRAFT_PROFILE_URL: &str = "https://api.minecraftservices.com/minecraft/profile";

#[derive(Debug, Deserialize)]
#[serde(rename_all = "PascalCase")]
struct XboxLiveResponse {
    token: String,
    /// ```json
    /// {"xui": [{"uhs": "userhash"}]}
    /// ```
    display_claims: serde_json::Value,
}

impl XboxLiveResponse {
    fn get_userhash(&self) -> Result<String> {
        self.display_claims
            .get("xui")
            .and_then(|v| v.get(0))
            .and_then(|v| v.get("uhs"))
            .and_then(|v| v.as_str())
            .map(std::borrow::ToOwned::to_owned)
            .ok_or_else(|| AuthenticationError::InvalidDisplayClaims(self.display_claims.clone()))
    }
}

#[derive(Debug, Deserialize)]
#[serde(rename_all = "PascalCase")]
pub struct XstsErrorResponse {
    x_err: u32,
    message: String,
    redirect: String,
}

#[derive(Debug, Deserialize)]
struct MinecraftLoginResponse {
    access_token: String,
}

#[derive(Debug, Deserialize)]
pub struct MinecraftProfile {
    pub id: String,
    pub name: String,
}

#[derive(Debug, Error)]
pub enum AuthenticationError {
    #[error("reqwest error")]
    Reqwest(#[from] reqwest::Error),
    /// from `eyre::Report` because oauth2 error types are too complicated
    #[error(transparent)]
    OAuth2(#[from] Report),
    #[error("userhash mismatch (xbox live: {0} / xsts: {1})")]
    UserHashMismatch(String, String),
    #[error("invalid display claims: {0}")]
    InvalidDisplayClaims(serde_json::Value),
    #[error("authentication failed: {0} returned {1}, body:\n{2}")]
    AuthenticationFailed(Url, StatusCode, String),
    #[error(transparent)]
    XstsError(#[from] XstsError),
}

impl AuthenticationError {
    async fn from_response(response: Response) -> Self {
        let url = response.url().clone();
        let status = response.status();

        match response.text().await {
            Ok(body) => Self::AuthenticationFailed(url, status, body),
            Err(e) => Self::Reqwest(e),
        }
    }
}

#[derive(Debug, Error)]
pub enum XstsError {
    #[error("this account does not have an xbox account")]
    NoXboxAccount,
    #[error("xbox live is unavailable in your region/country")]
    XboxLiveUnavailable,
    #[error("the account needs adult verification (south korea)")]
    AdultVerificationNeededSK,
    #[error("the account is a child and needs to be added to a family")]
    AddToFamily,
    #[error("unknown error: {}, {} (see {})", .0.x_err, .0.message, .0.redirect)]
    Unknown(XstsErrorResponse),
}

impl From<XstsErrorResponse> for XstsError {
    #[allow(clippy::unreadable_literal)]
    fn from(res: XstsErrorResponse) -> Self {
        match res.x_err {
            2148916233 => Self::NoXboxAccount,
            2148916235 => Self::XboxLiveUnavailable,
            2148916236 | 2148916237 => Self::AdultVerificationNeededSK,
            2148916238 => Self::AddToFamily,
            _ => Self::Unknown(res),
        }
    }
}

type Result<T> = std::result::Result<T, AuthenticationError>;

// TODO better error handling !!!!
pub async fn login() -> Result<MinecraftProfile> {
    // temporary until i can access api.minecraftservices.com

    let access_token = request_msa_token().await?;
    let (xbox_token, userhash) = auth_xbox_live(&access_token).await?;
    let (xsts_token, xsts_userhash) = fetch_xsts_token(&xbox_token).await?;

    if userhash != xsts_userhash {
        return Err(AuthenticationError::UserHashMismatch(
            userhash,
            xsts_userhash,
        ));
    }

    let mc_token = login_to_mc(&xsts_token, &userhash).await?;

    get_minecraft_profile(&mc_token).await
}

async fn request_msa_token() -> Result<String> {
    let client = BasicClient::new(
        ClientId::new(CLIENT_ID.to_string()),
        None,
        AuthUrl::new(MSA_AUTHORIZE_URL.to_string()).map_err(Report::from)?,
        Some(TokenUrl::new(MSA_TOKEN_URL.to_string()).map_err(Report::from)?),
    )
    .set_device_authorization_url(
        DeviceAuthorizationUrl::new(DEVICE_CODE_URL.to_string()).map_err(Report::from)?,
    );

    let details: StandardDeviceAuthorizationResponse = client
        .exchange_device_code()
        .map_err(Report::from)?
        .add_scope(Scope::new("XboxLive.signin offline_access".to_string()))
        .request_async(async_http_client)
        .await
        .map_err(Report::from)?;

    tracing::info!(
        "url: {} code: {}",
        **details.verification_uri(),
        details.user_code().secret()
    );

    let token_result = client
        .exchange_device_access_token(&details)
        .request_async(
            async_http_client,
            tokio::time::sleep,
            Some(Duration::from_secs(300)),
        )
        .await
        .map_err(Report::from)?;

    tracing::info!("fetched msa account token");

    Ok(token_result.access_token().secret().clone())
}

async fn auth_xbox_live(msa_token: &str) -> Result<(String, String)> {
    tracing::info!("fetching xbox token");

    let json = json!({
        "Properties": {
            "AuthMethod": "RPS",
            "SiteName": "user.auth.xboxlive.com",
            "RpsTicket": format!("d={msa_token}") // idek man
        },
        "RelyingParty": "http://auth.xboxlive.com",
        "TokenType": "JWT"
    });

    let response = crate::HTTP_CLIENT
        .post(XBOX_USER_AUTH_URL)
        .json(&json)
        .send()
        .await?;

    if response.status() == StatusCode::OK {
        let xbox_res = response.json::<XboxLiveResponse>().await?;
        let userhash = xbox_res.get_userhash()?;

        Ok((xbox_res.token, userhash))
    } else {
        Err(AuthenticationError::from_response(response).await)
    }
}

async fn fetch_xsts_token(xbox_token: &str) -> Result<(String, String)> {
    tracing::info!("fetching xbox xsts token");

    let json = json!({
        "Properties": {
            "SandboxId": "RETAIL",
            "UserTokens": [xbox_token]
        },
        "RelyingParty": "rp://api.minecraftservices.com/",
        "TokenType": "JWT"
    });

    let response = crate::HTTP_CLIENT
        .post(XBOX_XSTS_AUTHORIZE_URL)
        .json(&json)
        .send()
        .await?;

    match response.status() {
        StatusCode::OK => {
            let xsts_res = response.json::<XboxLiveResponse>().await?;
            let userhash = xsts_res.get_userhash()?;

            Ok((xsts_res.token, userhash))
        }
        StatusCode::UNAUTHORIZED => {
            let xsts_err = response.json::<XstsErrorResponse>().await?;

            Err(XstsError::from(xsts_err).into())
        }
        _ => Err(AuthenticationError::from_response(response).await),
    }
}

async fn login_to_mc(xsts_token: &str, userhash: &str) -> Result<String> {
    tracing::info!("logging in to minecraft");

    let json = json!({
        "identityToken": format!("XBL3.0 x={userhash};{xsts_token}")
    });

    let response = crate::HTTP_CLIENT
        .post(MINECRAFT_AUTH_URL)
        .json(&json)
        .send()
        .await?;

    if response.status() == StatusCode::OK {
        let mc_res = response.json::<MinecraftLoginResponse>().await?;

        Ok(mc_res.access_token)
    } else {
        Err(AuthenticationError::from_response(response).await)
    }
}

async fn get_minecraft_profile(mc_token: &str) -> Result<MinecraftProfile> {
    tracing::info!("fetching player profile");

    crate::HTTP_CLIENT
        .get(MINECRAFT_PROFILE_URL)
        .bearer_auth(mc_token)
        .send()
        .await?
        .json::<MinecraftProfile>()
        .await
        .map_err(AuthenticationError::from)
}
