use std::{env, fs::File, path::PathBuf};

use chrono::{DateTime, Utc};
use color_eyre::eyre::{bail, Report, Result};
use itertools::Itertools;
use serde::Deserialize;
use sha2::{Digest, Sha512};

const PROJECT_ID: &str = "JR0bkFKa";

#[derive(Debug, Deserialize)]
pub struct ProjectVersion {
    pub name: String,
    pub version_number: String,
    game_versions: Vec<String>,
    files: Vec<VersionFile>,
    date_published: DateTime<Utc>,
}

#[derive(Debug, Deserialize)]
pub struct VersionFile {
    url: String,
    filename: String,
    size: usize,
    hashes: VersionFileHashes,
}

#[derive(Debug, Deserialize)]
pub struct VersionFileHashes {
    sha512: String,
}

impl ProjectVersion {
    pub async fn download_file(&self) -> Result<PathBuf> {
        let version_file = &self.files[0];
        let path = env::temp_dir().join(&version_file.filename);

        let response = crate::HTTP_CLIENT.get(&version_file.url).send().await?;
        let mut bytes = &*response.bytes().await?; // deref to get &[u8]

        if bytes.len() != version_file.size {
            bail!("invalid length!")
        }

        let expected_hash = hex::decode(&version_file.hashes.sha512)?;
        let computed_hash = {
            let mut sha512 = Sha512::new();
            sha512.update(bytes);
            sha512.finalize()
        };

        if expected_hash[..] != computed_hash[..] {
            bail!("hash mismatch!")
        }

        let mut file = File::create(&path)?;
        std::io::copy(&mut bytes, &mut file)?;

        Ok(path)
    }
}

pub async fn get_versions() -> Result<Vec<ProjectVersion>> {
    let url = format!("https://api.modrinth.com/v2/project/{PROJECT_ID}/version");

    crate::HTTP_CLIENT
        .get(url)
        .send()
        .await?
        .json()
        .await
        .map_err(Report::from)
}

pub trait VersionsExt {
    fn available_minecraft_versions(&self) -> Vec<&String>;
    fn latest_for_mc_version(&self, mc_version: &str) -> Option<&ProjectVersion>;
}

impl VersionsExt for Vec<ProjectVersion> {
    fn available_minecraft_versions(&self) -> Vec<&String> {
        // we can safely assume that there is only one game_version per project version since it's a modpack
        self.iter().map(|v| &v.game_versions[0]).unique().collect()
    }

    fn latest_for_mc_version(&self, mc_version: &str) -> Option<&ProjectVersion> {
        self.iter()
            .filter(|v| v.game_versions.contains(&mc_version.to_owned()))
            .sorted_by_key(|v| v.date_published)
            .next_back()
    }
}
